/*
        
        File:			AtheroJackDriver.cpp
        Program:		AtheroJack
	Author:			Michael Rossberg
				mick@binaervarianz.de
	Description:		KisMAC is a wireless stumbler for MacOS X.
                
        This file is part of KisMAC.

    KisMAC is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    KisMAC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KisMAC; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <IOKit/IODeviceTreeSupport.h>
#include "AtheroJackDriver.h"
#include "AtheroJackUserClient.h"
#include <IOKit/IOService.h>
#include <IOKit/IOMessage.h>
#include <IOKit/pccard/IOPCCard.h>
#include <IOKit/pccard/IOCardBusDevice.h>

#define super IOService

OSDefineMetaClassAndStructors(AtheroJackDriver, IOService);

bool AtheroJackDriver::init(OSDictionary* properties) {
    WLLogInfo("AtheroJackDriver::init: start\n");
    
    if (super::init(properties) == false) {
	WLLogErr("AtheroJackDriver::init: super::init failed\n");
	return false;
    }
    _wlCard = NULL;
    _map = NULL;
    
    return true;
}

/*
 * The start() function is responsible for setting up the resources
 * the driver needs all the time, whether the driver is enabled or
 * disabled.  These include the network interface object, a work loop,
 * and an output queue, along with whatever specific resources the
 * driver needs.
 */
bool AtheroJackDriver::start(IOService* provider) {
    WLLogInfo("AtheroJackDriver::start: start\n");
    
    if (!super::start(provider)) {
	WLLogErr("AtheroJackDriver::start: super::start failed\n");
	return false;
    }

    /*
     * Verify that the provider is non-null and is the appropriate
     * class.
     *
     * XXX: AppleMacIODevice is valid for the AirPort, does it work
     * PC cards too?
     */
    _nub = OSDynamicCast(IOPCIDevice, provider);
    if (!_nub) {
	WLLogCrit("AtheroJackDriver::start: Null nub\n");
	return false;
    }
    
    // Open provider
    if (!_openProvider()) {
	WLLogErr("AtheroJackDriver::start: Couldn't open provider\n");
	return false;
    }

    /*
     * From this point on, we must close the provider if something
     * fails and we abort.
     */
    bool ret = true;
    do {
        WLLogDebug("AtheroJackDriver::start: Configuring the card\n");
        
        int alignSize = _nub->configRead8(kIOPCIConfigCacheLineSize);
        alignSize = alignSize;    // max the 8-bit latency count
        
        /*
	 * Get a memory map and virtual base address and use the base
	 * address to instantiate a WLCard.  The WLCard will
	 * figure out what card/vendor/revision it is and act
	 * accordingly.
	 */
	WLLogDebug("AtheroJackDriver::start: Mapping the card\n");
        _map = _nub->mapDeviceMemoryWithIndex(0);
	if (!_map) {
	    WLLogErr("AtheroJackDriver::start: Couldn't map device memory\n");
	    ret = false;
	    break;
	}
	
        WLLogDebug("AtheroJackDriver::start: Getting the virtual address of the card\n");
        IOVirtualAddress ioBase = _map->getVirtualAddress();
        if (ioBase==0) {
            WLLogErr("AtheroJackDriver::start: Got an empty ioBase!\n");
	    ret = false;
	    break;
        }

        //enable memory response from the card
        _nub->setMemoryEnable(true);
        
	/*
	 * Instantiate Atheros card
	 */
        WLLogDebug("AtheroJackDriver::start: Start the card\n");
	_wlCard = new AtheroJackCard((void*)ioBase, NULL, this, alignSize);

        /*
         * Create a packet queue
         */
        UInt32 pqSize = 16 * 1024;
        _packetQueue = IODataQueue::withCapacity(pqSize);
        
	/*
	 * Initialize work loop and event sources
	 */
	ret = _initEventSources();
	if (!ret)
	    break;

        if (_workLoop) _workLoop->enableAllInterrupts();
        /*
         * Register driver in registry
         */	
	registerService();
    } while (0);

    if (!ret) {
        if (_map) _map->release();
        _closeProvider();
    }
    return ret;
}

bool AtheroJackDriver::_initEventSources() {
    /*
     * Get our work loop and add interrupt, watchdog timer, and user
     * client command event sources to it.
     */
    IOWorkLoop* wl = getWorkLoop();
    if (!wl) {
	WLLogErr("AtheroJackDriver::start: Couldn't get IOWorkLoop\n");
	return false;
    }    

    /* Create an interrupt event source */
    WLLogDebug("AtheroJackDriver::start: Register Interupt source\n");
    _interruptEventSource = IOInterruptEventSource::interruptEventSource(this, 
        (IOInterruptEventAction)&AtheroJackDriver::_interruptOccurred, _nub, 0);
    
    if (!_interruptEventSource) {
	WLLogErr("AtheroJackDriver::start: Couldn't get interrupt event source\n");
	return false;
    }
    else if (wl->addEventSource(_interruptEventSource) != kIOReturnSuccess) {
	WLLogErr("AtheroJackDriver::start: Couldn't add interrupt event source "
                 "to work loop\n");
	return false;
    }
    _interruptEventSource->enable();
    
    /* Create a timer event source */
    _timerEventSource = IOTimerEventSource::
	timerEventSource(this, (IOTimerEventSource::Action)
			 &AtheroJackDriver::_timeoutOccurred);
    if (!_timerEventSource) {
	WLLogErr("AtheroJackDriver::start: Couldn't get timer event source\n");
	return false;
    }
    else if (wl->addEventSource(_timerEventSource) != kIOReturnSuccess) {
	WLLogErr("AtheroJackDriver::start: Couldn't add timer event source "
                 "to work loop\n");
	return false;
    }
    
    /* Create a command gate for user client commands */
    _userCommandGate = IOCommandGate::commandGate(this);
    if (!_userCommandGate) {
	WLLogErr("AtheroJackDriver::start: Couldn't get command gate\n");
	return false;
    }
    else if (wl->addEventSource(_userCommandGate) != kIOReturnSuccess) {
	WLLogErr("AtheroJackDriver::start: Couldn't add cmd gate to work loop\n");
	return false;
    }

    return true;
}

/*
 * The stop() function simply reverses all the actions taken by
 * start(), releasing the nub and disposing of any resources created
 * in start().
 */
void AtheroJackDriver::stop(IOService* provider) {
    WLLogInfo("AtheroJackDriver::stop()\n");

    if (_workLoop) _workLoop->disableAllInterrupts();
    
    if (_userCommandGate) {
        getWorkLoop()->removeEventSource(_userCommandGate);
        _userCommandGate->release();
        _userCommandGate = 0;
    }
    
    if (_wlCard) {
        delete _wlCard;
        _wlCard = 0;
    }
    
    if (_timerEventSource) {
	_timerEventSource->disable();
	getWorkLoop()->removeEventSource(_timerEventSource);
	_timerEventSource->release();
	_timerEventSource = 0;
    }

    if (_interruptEventSource) {
	_interruptEventSource->disable();
        getWorkLoop()->removeEventSource(_interruptEventSource);
        _interruptEventSource->release();
        _interruptEventSource = 0;
    }
    
    if (_workLoop) {
        _workLoop->release();
        _workLoop = 0;
    }
    
    if (_nub) {
        //_nub->unconfigure();
        //_nub->close(this);
    }    

    if (_map) _map->release();
    
    _closeProvider();
    
    super::stop(provider);
}

void AtheroJackDriver::free() {
    WLLogInfo("AtheroJackDriver::free()\n");
    super::free();
}

IOWorkLoop* AtheroJackDriver::getWorkLoop() {
    if (!_workLoop)
        _workLoop = IOWorkLoop::workLoop();
        
    return _workLoop;
}

IOReturn AtheroJackDriver::message(UInt32 type, IOService * provider, void * argument) {
    if (type == kIOPCCardCSEventMessage) {
	WLLogInfo("AtheroJackDriver::message: nub=%p, CS event received type=0x%x.\n", _nub, (unsigned int)argument);

        switch ((unsigned int)argument) {
	    case CS_EVENT_CARD_REMOVAL:
		WLLogInfo("AtheroJackDriver::message: PC Card was removed\n");
                
                if (_wlCard) _wlCard->cardGone();
		
		// Message our clients and tell them we're no longer available...
		super::message(kIOMessageServiceIsTerminated, this, 0);                
		break;
            case CS_EVENT_CARD_INSERTION:
                WLLogInfo("AtheroJackDriver::message: PC Card was inserted. \n");
                //TODO Handle this              
		break;
	    default:
		WLLogInfo("AtheroJackDriver::message: PC Card generated untrapped message: %ud\n", (unsigned int)argument);
		break;
        }
    }

    return super::message(type, provider, argument);
}

bool AtheroJackDriver::
handleOpen(IOService* forClient, IOOptionBits options, void* arg) {
    if (!super::handleOpen(forClient, options, arg)) {
	WLLogErr("AtheroJackDriver::handleOpen: super::handleOpen failed.\n");
	return false;
    }

    if (!_openProvider()) {
	WLLogErr("AtheroJackDriver::handleOpen: _openProvider failed\n");
	super::handleClose(forClient, 0);
	return false;
    }

    WLLogInfo("AtheroJackDriver::handleOpen\n");

    return true;
}

void AtheroJackDriver::
handleClose(IOService* forClient, IOOptionBits options) {
    WLLogInfo("AtheroJackDriver::handleClose\n");
    _closeProvider();
    super::handleClose(forClient, options);
}

bool AtheroJackDriver::
_openProvider() {
    if (OSIncrementAtomic(&_openCount) > 0)
        return true;
        
    if (!_nub->open(this)) {
        WLLogErr("AtheroJackDriver::_openProvider: Couldn't open nub\n");
        return false;
    }
    
    return true;
}

void AtheroJackDriver::
_closeProvider() {
    if (OSDecrementAtomic(&_openCount) > 1)
	return;
    _nub->close(this);
}

/*
 * Static event-handlers just call the virtual handler functions
 */
void AtheroJackDriver::
_interruptOccurred(OSObject* o, IOInterruptEventSource* src, int n) {
    ((AtheroJackDriver*)o)->handleInterrupt(src, n);
}

void AtheroJackDriver::
_timeoutOccurred(OSObject* o, IOTimerEventSource* src) {
    ((AtheroJackDriver*)o)->handleTimeout(src);
}

void AtheroJackDriver::
handleInterrupt(IOInterruptEventSource* eventSource, int count) {
    if (_wlCard) _wlCard->handleInterrupt(); 
}

void AtheroJackDriver::
handleTimeout(IOTimerEventSource* eventSource) {
    WLLogInfo("AtheroJackDriver::handleTimeout()\n");
}


/*********************************************************************
 *                         Power Management                          *
 *********************************************************************/

IOReturn AtheroJackDriver::
powerStateWillChangeTo(IOPMPowerFlags capabilities,
		      unsigned long stateNumber,
		      IOService* whatDevice) {
    WLLogNotice("AtheroJackDriver::powerStateWillChangeTo: %ld\n", stateNumber);

    _nub->setPowerState(stateNumber,whatDevice);
    
    return IOPMAckImplied;
}

IOReturn AtheroJackDriver::
powerStateDidChangeTo(IOPMPowerFlags capabilities,
		      unsigned long stateNumber,
		      IOService* whatDevice) {
    WLLogNotice("AtheroJackDriver::powerStateDidChangeTo: %ld\n", stateNumber);

    if (!_wlCard) return IOPMAckImplied;
    
    if (stateNumber == kIOPCIDeviceOnState) {
        IOSleep(100);
        //_wlCard->wakeUp();
        //_wlCard->_reset();
    } else {
        //_wlCard->goSleep();
    }

    return IOPMAckImplied;
}
